import types
import typing
from collections.abc import Callable, Sequence
from typing import Any

from haiway import types as _types
from haiway.state.attributes import AttributeAnnotation

__all__ = [
    "attribute_type_validator",
]


def attribute_type_validator(
    annotation: AttributeAnnotation,
    /,
) -> Callable[[Any], Any]:
    match annotation.origin:
        case types.NoneType:
            return _none_validator

        case _types.Missing:
            return _missing_validator

        case types.UnionType:
            return _prepare_union_validator(annotation.arguments)

        case typing.Literal:
            return _prepare_literal_validator(annotation.arguments)

        case typing.Any:
            return _any_validator

        case type() as other_type:
            return _prepare_type_validator(other_type)

        case other:
            raise TypeError(f"Unsupported type annotation: {other}")


def _none_validator(
    value: Any,
) -> Any:
    match value:
        case None:
            return None

        case _:
            raise TypeError(f"Type '{type(value)}' is not matching expected type 'None'")


def _missing_validator(
    value: Any,
) -> Any:
    match value:
        case _types.Missing():
            return _types.MISSING

        case _:
            raise TypeError(f"Type '{type(value)}' is not matching expected type 'Missing'")


def _any_validator(
    value: Any,
) -> Any:
    return value  # any is always valid


def _prepare_union_validator(
    elements: Sequence[AttributeAnnotation],
    /,
) -> Callable[[Any], Any]:
    validators: list[Callable[[Any], Any]] = [
        attribute_type_validator(alternative) for alternative in elements
    ]

    def union_validator(
        value: Any,
    ) -> Any:
        errors: list[Exception] = []
        for validator in validators:
            try:
                return validator(value)

            except Exception as exc:
                errors.append(exc)

        raise ExceptionGroup("Multiple errors", errors)

    return union_validator


def _prepare_literal_validator(
    elements: Sequence[Any],
    /,
) -> Callable[[Any], Any]:
    def literal_validator(
        value: Any,
    ) -> Any:
        if value in elements:
            return value

        else:
            raise ValueError(f"Value '{value}' is not matching expected '{elements}'")

    return literal_validator


def _prepare_type_validator(
    validated_type: type[Any],
    /,
) -> Callable[[Any], Any]:
    def type_validator(
        value: Any,
    ) -> Any:
        match value:
            case value if isinstance(value, validated_type):
                return value

            case _:
                raise TypeError(
                    f"Type '{type(value)}' is not matching expected type '{validated_type}'"
                )

    return type_validator



from haiway.utils.always import always, async_always
from haiway.utils.env import getenv_bool, getenv_float, getenv_int, getenv_str, load_env
from haiway.utils.immutable import freeze
from haiway.utils.logs import setup_logging
from haiway.utils.mimic import mimic_function
from haiway.utils.noop import async_noop, noop
from haiway.utils.queue import AsyncQueue

__all__ = [
    "always",
    "async_always",
    "async_noop",
    "AsyncQueue",
    "freeze",
    "getenv_bool",
    "getenv_float",
    "getenv_int",
    "getenv_str",
    "load_env",
    "mimic_function",
    "noop",
    "setup_logging",
]



from collections.abc import Callable
from typing import Literal, Protocol, Self, runtime_checkable

from haiway import Structure, frozenlist


def test_basic_initializes_with_arguments() -> None:
    @runtime_checkable
    class Proto(Protocol):
        def __call__(self) -> None: ...

    class Basics(Structure):
        string: str
        literal: Literal["A", "B"]
        sequence: list[str]
        frozen: frozenlist[int]
        integer: int
        union: str | int
        optional: str | None
        none: None
        function: Callable[[], None]
        proto: Proto

    basic = Basics(
        string="string",
        literal="A",
        sequence=["a", "b", "c"],
        frozen=(1, 2, 3),
        integer=0,
        union="union",
        optional="optional",
        none=None,
        function=lambda: None,
        proto=lambda: None,
    )
    assert basic.string == "string"
    assert basic.literal == "A"
    assert basic.sequence == ["a", "b", "c"]
    assert basic.frozen == (1, 2, 3)
    assert basic.integer == 0
    assert basic.union == "union"
    assert basic.optional == "optional"
    assert basic.none is None
    assert callable(basic.function)
    assert isinstance(basic.proto, Proto)


def test_basic_initializes_with_defaults() -> None:
    class Basics(Structure):
        string: str = ""
        integer: int = 0
        optional: str | None = None

    basic = Basics()
    assert basic.string == ""
    assert basic.integer == 0
    assert basic.optional is None


def test_basic_initializes_with_arguments_and_defaults() -> None:
    class Basics(Structure):
        string: str
        integer: int = 0
        optional: str | None = None

    basic = Basics(
        string="string",
        integer=42,
    )
    assert basic.string == "string"
    assert basic.integer == 42
    assert basic.optional is None


def test_parametrized_initializes_with_proper_parameters() -> None:
    class Parametrized[T](Structure):
        value: T

    parametrized_string = Parametrized(
        value="string",
    )
    assert parametrized_string.value == "string"

    parametrized_int = Parametrized(
        value=42,
    )
    assert parametrized_int.value == 42

    assert parametrized_string != parametrized_int


def test_nested_initializes_with_proper_arguments() -> None:
    class Nested(Structure):
        string: str

    class Recursive(Structure):
        nested: Nested
        recursion: "Recursive | None"
        self_recursion: Self | None

    recursive = Recursive(
        nested=Nested(string="one"),
        recursion=Recursive(
            nested=Nested(string="two"),
            recursion=None,
            self_recursion=None,
        ),
        self_recursion=None,
    )
    assert recursive.nested == Nested(string="one")
    assert recursive.recursion == Recursive(
        nested=Nested(string="two"),
        recursion=None,
        self_recursion=None,
    )



from abc import ABC, abstractmethod
from asyncio import Lock, gather, shield
from typing import ClassVar, Self, cast, final

__all__ = [
    "Dependencies",
    "Dependency",
]


class Dependency(ABC):
    @classmethod
    @abstractmethod
    async def prepare(cls) -> Self: ...

    async def dispose(self) -> None:  # noqa: B027
        pass


@final
class Dependencies:
    _lock: ClassVar[Lock] = Lock()
    _dependencies: ClassVar[dict[type[Dependency], Dependency]] = {}

    def __init__(self) -> None:
        raise NotImplementedError("Can't instantiate Dependencies")

    @classmethod
    async def dependency[Requested: Dependency](
        cls,
        dependency: type[Requested],
        /,
    ) -> Requested:
        async with cls._lock:
            if dependency not in cls._dependencies:
                cls._dependencies[dependency] = await dependency.prepare()

            return cast(Requested, cls._dependencies[dependency])

    @classmethod
    async def register(
        cls,
        dependency: Dependency,
        /,
    ) -> None:
        async with cls._lock:
            if current := cls._dependencies.get(dependency.__class__):
                await current.dispose()

            cls._dependencies[dependency.__class__] = dependency

    @classmethod
    async def dispose(cls) -> None:
        async with cls._lock:
            await shield(
                gather(
                    *[dependency.dispose() for dependency in cls._dependencies.values()],
                    return_exceptions=False,
                )
            )
            cls._dependencies.clear()



from haiway.helpers.asynchronous import asynchronous
from haiway.helpers.cache import cached
from haiway.helpers.retry import auto_retry
from haiway.helpers.throttling import throttle
from haiway.helpers.timeout import with_timeout

__all__ = [
    "asynchronous",
    "auto_retry",
    "cached",
    "throttle",
    "with_timeout",
]



__all__ = [
    "MissingContext",
    "MissingDependency",
    "MissingState",
]


class MissingContext(Exception):
    pass


class MissingDependency(Exception):
    pass


class MissingState(Exception):
    pass



from haiway import MissingContext, ScopeMetrics, Structure, ctx
from pytest import mark, raises


class ExampleState(Structure):
    state: str = "default"


class FakeException(Exception):
    pass


@mark.asyncio
async def test_state_is_available_according_to_context():
    with raises(MissingContext):
        assert ctx.state(ExampleState).state == "default"

    async with ctx.scope("default"):
        assert ctx.state(ExampleState).state == "default"

        async with ctx.scope("specified", ExampleState(state="specified")):
            assert ctx.state(ExampleState).state == "specified"

            async with ctx.scope("modified", ExampleState(state="modified")):
                assert ctx.state(ExampleState).state == "modified"

            assert ctx.state(ExampleState).state == "specified"

        assert ctx.state(ExampleState).state == "default"

    with raises(MissingContext):
        assert ctx.state(ExampleState).state == "default"


@mark.asyncio
async def test_state_update_updates_local_context():
    with raises(MissingContext):
        assert ctx.state(ExampleState).state == "default"

    async with ctx.scope("default"):
        assert ctx.state(ExampleState).state == "default"

        with ctx.updated(ExampleState(state="updated")):
            assert ctx.state(ExampleState).state == "updated"

            with ctx.updated(ExampleState(state="modified")):
                assert ctx.state(ExampleState).state == "modified"

            assert ctx.state(ExampleState).state == "updated"

        assert ctx.state(ExampleState).state == "default"

    with raises(MissingContext):
        assert ctx.state(ExampleState).state == "default"


@mark.asyncio
async def test_exceptions_are_propagated():
    with raises(FakeException):
        async with ctx.scope("outer"):
            async with ctx.scope("inner"):
                raise FakeException()


@mark.asyncio
async def test_completions_are_called_according_to_context_exits():
    executions: int = 0

    async def completion(metrics: ScopeMetrics):
        nonlocal executions
        executions += 1

    async with ctx.scope("outer", completion=completion):
        assert executions == 0

        async with ctx.scope("inner", completion=completion):
            assert executions == 0

        assert executions == 1

    assert executions == 2


@mark.asyncio
async def test_metrics_are_recorded_within_context():
    def verify_example_metrics(state: str):
        async def completion(metrics: ScopeMetrics):
            assert metrics.read(ExampleState, default=ExampleState()).state == state

        return completion

    async with ctx.scope("outer", completion=verify_example_metrics("outer-in-out")):
        ctx.record(ExampleState(state="outer-in"))

        async with ctx.scope("inner", completion=verify_example_metrics("inner")):
            ctx.record(ExampleState(state="inner"))

        ctx.record(
            ExampleState(state="-out"),
            merge=lambda lhs, rhs: ExampleState(state=lhs.state + rhs.state),
        )



from asyncio import Task, TaskGroup, get_event_loop
from collections.abc import Callable, Coroutine
from contextvars import ContextVar, Token, copy_context
from types import TracebackType
from typing import final

__all__ = [
    "TaskGroupContext",
]


@final
class TaskGroupContext:
    _context = ContextVar[TaskGroup]("TaskGroupContext")

    @classmethod
    def run[Result, **Arguments](
        cls,
        function: Callable[Arguments, Coroutine[None, None, Result]],
        /,
        *args: Arguments.args,
        **kwargs: Arguments.kwargs,
    ) -> Task[Result]:
        try:
            return cls._context.get().create_task(
                function(*args, **kwargs),
                context=copy_context(),
            )

        except LookupError:  # spawn task out of group as a fallback
            return get_event_loop().create_task(
                function(*args, **kwargs),
                context=copy_context(),
            )

    def __init__(
        self,
    ) -> None:
        self._group: TaskGroup = TaskGroup()
        self._token: Token[TaskGroup] | None = None

    async def __aenter__(self) -> None:
        assert self._token is None, "TaskGroupContext reentrance is not allowed"  # nosec: B101
        await self._group.__aenter__()
        self._token = TaskGroupContext._context.set(self._group)

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._token is not None, "Unbalanced TaskGroupContext context exit"  # nosec: B101
        TaskGroupContext._context.reset(self._token)
        self._token = None

        try:
            await self._group.__aexit__(
                et=exc_type,
                exc=exc_val,
                tb=exc_tb,
            )

        except BaseException:
            pass  # silence TaskGroup exceptions, if there was exception already we will get it



from collections.abc import Iterable
from contextvars import ContextVar, Token
from types import TracebackType
from typing import Self, cast, final

from haiway.context.types import MissingContext, MissingState
from haiway.state import Structure
from haiway.utils import freeze

__all__ = [
    "ScopeState",
    "StateContext",
]


@final
class ScopeState:
    def __init__(
        self,
        state: Iterable[Structure],
    ) -> None:
        self._state: dict[type[Structure], Structure] = {
            type(element): element for element in state
        }
        freeze(self)

    def state[State: Structure](
        self,
        state: type[State],
        /,
        default: State | None = None,
    ) -> State:
        if state in self._state:
            return cast(State, self._state[state])

        elif default is not None:
            return default

        else:
            try:
                initialized: State = state()
                self._state[state] = initialized
                return initialized

            except Exception as exc:
                raise MissingState(
                    f"{state.__qualname__} is not defined in current scope"
                    " and failed to provide a default value"
                ) from exc

    def updated(
        self,
        state: Iterable[Structure],
    ) -> Self:
        if state:
            return self.__class__(
                [
                    *self._state.values(),
                    *state,
                ]
            )

        else:
            return self


@final
class StateContext:
    _context = ContextVar[ScopeState]("StateContext")

    @classmethod
    def current[State: Structure](
        cls,
        state: type[State],
        /,
        default: State | None = None,
    ) -> State:
        try:
            return cls._context.get().state(state, default=default)

        except LookupError as exc:
            raise MissingContext("StateContext requested but not defined!") from exc

    @classmethod
    def updated(
        cls,
        state: Iterable[Structure],
        /,
    ) -> Self:
        try:
            return cls(state=cls._context.get().updated(state=state))

        except LookupError:  # create new context as a fallback
            return cls(state=ScopeState(state))

    def __init__(
        self,
        state: ScopeState,
    ) -> None:
        self._state: ScopeState = state
        self._token: Token[ScopeState] | None = None

    def __enter__(self) -> None:
        assert self._token is None, "StateContext reentrance is not allowed"  # nosec: B101
        self._token = StateContext._context.set(self._state)

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert self._token is not None, "Unbalanced StateContext context exit"  # nosec: B101
        StateContext._context.reset(self._token)
        self._token = None



from haiway.context.access import ctx
from haiway.context.dependencies import Dependencies, Dependency
from haiway.context.metrics import ScopeMetrics
from haiway.context.types import MissingContext, MissingDependency, MissingState

__all__ = [
    "ctx",
    "Dependencies",
    "Dependency",
    "MissingContext",
    "MissingDependency",
    "MissingState",
    "ScopeMetrics",
]



import sys
import types
import typing
from collections.abc import Mapping
from types import NoneType, UnionType
from typing import (
    Any,
    ClassVar,
    ForwardRef,
    Generic,
    Literal,
    TypeAliasType,
    TypeVar,
    get_args,
    get_origin,
    get_type_hints,
)

__all__ = [
    "attribute_annotations",
    "AttributeAnnotation",
]


class AttributeAnnotation:
    def __init__(
        self,
        *,
        origin: Any,
        arguments: list[Any],
    ) -> None:
        self.origin: Any = origin
        self.arguments: list[Any] = arguments

    def __eq__(
        self,
        other: Any,
    ) -> bool:
        return self is other or (
            isinstance(other, self.__class__)
            and self.origin == other.origin
            and self.arguments == other.arguments
        )


def attribute_annotations(
    cls: type[Any],
    /,
    type_parameters: dict[str, Any] | None = None,
) -> dict[str, AttributeAnnotation]:
    type_parameters = type_parameters or {}

    self_annotation = AttributeAnnotation(
        origin=cls,
        arguments=[],  # ignore self arguments here, Structure will have them resolved at this stage
    )
    localns: dict[str, Any] = {cls.__name__: cls}
    recursion_guard: dict[Any, AttributeAnnotation] = {cls: self_annotation}
    attributes: dict[str, AttributeAnnotation] = {}

    for key, annotation in get_type_hints(cls, localns=localns).items():
        # do not include ClassVars, private or dunder items
        if ((get_origin(annotation) or annotation) is ClassVar) or key.startswith("_"):
            continue

        attributes[key] = _resolve_attribute_annotation(
            annotation,
            self_annotation=self_annotation,
            type_parameters=type_parameters,
            module=cls.__module__,
            localns=localns,
            recursion_guard=recursion_guard,
        )

    return attributes


def _resolve_attribute_annotation(  # noqa: C901, PLR0911, PLR0912, PLR0913
    annotation: Any,
    /,
    self_annotation: AttributeAnnotation | None,
    type_parameters: dict[str, Any],
    module: str,
    localns: dict[str, Any],
    recursion_guard: Mapping[Any, AttributeAnnotation],  # TODO: verify recursion!
) -> AttributeAnnotation:
    # resolve annotation directly if able
    match annotation:
        # None
        case types.NoneType | types.NoneType():
            return AttributeAnnotation(
                origin=NoneType,
                arguments=[],
            )

        # forward reference through string
        case str() as forward_ref:
            return _resolve_attribute_annotation(
                ForwardRef(forward_ref, module=module)._evaluate(
                    globalns=None,
                    localns=localns,
                    recursive_guard=frozenset(),
                ),
                self_annotation=self_annotation,
                type_parameters=type_parameters,
                module=module,
                localns=localns,
                recursion_guard=recursion_guard,  # we might need to update it somehow?
            )

        # forward reference directly
        case typing.ForwardRef() as reference:
            return _resolve_attribute_annotation(
                reference._evaluate(
                    globalns=None,
                    localns=localns,
                    recursive_guard=frozenset(),
                ),
                self_annotation=self_annotation,
                type_parameters=type_parameters,
                module=module,
                localns=localns,
                recursion_guard=recursion_guard,  # we might need to update it somehow?
            )

        # generic alias aka parametrized type
        case types.GenericAlias() as generic_alias:
            match get_origin(generic_alias):
                # check for an alias with parameters
                case typing.TypeAliasType() as alias:  # pyright: ignore[reportUnnecessaryComparison]
                    type_alias: AttributeAnnotation = AttributeAnnotation(
                        origin=TypeAliasType,
                        arguments=[],
                    )
                    resolved: AttributeAnnotation = _resolve_attribute_annotation(
                        alias.__value__,
                        self_annotation=None,
                        type_parameters=type_parameters,
                        module=module,
                        localns=localns,
                        recursion_guard=recursion_guard,
                    )
                    type_alias.origin = resolved.origin
                    type_alias.arguments = resolved.arguments
                    return type_alias

                # check if we can resolve it as generic
                case parametrized if issubclass(parametrized, Generic):
                    parametrized_type: Any = parametrized.__class_getitem__(  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
                        *(
                            type_parameters.get(
                                arg.__name__,
                                arg.__bound__ or Any,
                            )
                            if isinstance(arg, TypeVar)
                            else arg
                            for arg in get_args(generic_alias)
                        )
                    )

                    match parametrized_type:
                        # verify if we got any specific type or generic alias again
                        case types.GenericAlias():
                            return AttributeAnnotation(
                                origin=parametrized,
                                arguments=[
                                    _resolve_attribute_annotation(
                                        argument,
                                        self_annotation=self_annotation,
                                        type_parameters=type_parameters,
                                        module=module,
                                        localns=localns,
                                        recursion_guard=recursion_guard,
                                    )
                                    for argument in get_args(generic_alias)
                                ],
                            )

                        # use resolved type if it is not an alias again
                        case _:
                            return AttributeAnnotation(
                                origin=parametrized_type,
                                arguments=[],
                            )

                # anything else - try to resolve a concrete type or use as is
                case origin:
                    return AttributeAnnotation(
                        origin=origin,
                        arguments=[
                            _resolve_attribute_annotation(
                                argument,
                                self_annotation=self_annotation,
                                type_parameters=type_parameters,
                                module=module,
                                localns=localns,
                                recursion_guard=recursion_guard,
                            )
                            for argument in get_args(generic_alias)
                        ],
                    )

        # type alias
        case typing.TypeAliasType() as alias:
            type_alias: AttributeAnnotation = AttributeAnnotation(
                origin=TypeAliasType,
                arguments=[],
            )
            resolved: AttributeAnnotation = _resolve_attribute_annotation(
                alias.__value__,
                self_annotation=None,
                type_parameters=type_parameters,
                module=module,
                localns=localns,
                recursion_guard=recursion_guard,
            )
            type_alias.origin = resolved.origin
            type_alias.arguments = resolved.arguments
            return type_alias

        # type parameter
        case typing.TypeVar():
            return _resolve_attribute_annotation(
                # try to resolve it from current parameters if able
                type_parameters.get(
                    annotation.__name__,
                    # use bound as default or Any otherwise
                    annotation.__bound__ or Any,
                ),
                self_annotation=None,
                type_parameters=type_parameters,
                module=module,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case typing.ParamSpec():
            sys.stderr.write(
                "ParamSpec is not supported for attribute annotations,"
                " ignoring with Any type - it might incorrectly validate types\n"
            )
            return AttributeAnnotation(
                origin=Any,
                arguments=[],
            )

        case typing.TypeVarTuple():
            sys.stderr.write(
                "TypeVarTuple is not supported for attribute annotations,"
                " ignoring with Any type - it might incorrectly validate types\n"
            )
            return AttributeAnnotation(
                origin=Any,
                arguments=[],
            )

        case _:
            pass  # proceed to resolving based on origin

    # resolve based on origin if any
    match get_origin(annotation) or annotation:
        case types.UnionType | typing.Union:
            return AttributeAnnotation(
                origin=UnionType,  # pyright: ignore[reportArgumentType]
                arguments=[
                    recursion_guard.get(
                        argument,
                        _resolve_attribute_annotation(
                            argument,
                            self_annotation=self_annotation,
                            type_parameters=type_parameters,
                            module=module,
                            localns=localns,
                            recursion_guard=recursion_guard,
                        ),
                    )
                    for argument in get_args(annotation)
                ],
            )

        case typing.Callable:  # pyright: ignore[reportUnknownMemberType, reportAttributeAccessIssue]
            return AttributeAnnotation(
                origin=typing.Callable,
                arguments=[
                    _resolve_attribute_annotation(
                        argument,
                        self_annotation=self_annotation,
                        type_parameters=type_parameters,
                        module=module,
                        localns=localns,
                        recursion_guard=recursion_guard,
                    )
                    for argument in get_args(annotation)
                ],
            )

        case typing.Self:  # pyright: ignore[reportUnknownMemberType]
            if not self_annotation:
                sys.stderr.write(
                    "Unresolved Self attribute annotation,"
                    " ignoring with Any type - it might incorrectly validate types\n"
                )
                return AttributeAnnotation(
                    origin=Any,
                    arguments=[],
                )

            return self_annotation

        # unwrap from irrelevant type wrappers
        case typing.Annotated | typing.Final | typing.Required | typing.NotRequired:
            return _resolve_attribute_annotation(
                get_args(annotation)[0],
                self_annotation=self_annotation,
                type_parameters=type_parameters,
                module=module,
                localns=localns,
                recursion_guard=recursion_guard,
            )

        case typing.Optional:  # optional is a Union[Value, None]
            return AttributeAnnotation(
                origin=UnionType,  # pyright: ignore[reportArgumentType]
                arguments=[
                    _resolve_attribute_annotation(
                        get_args(annotation)[0],
                        self_annotation=self_annotation,
                        type_parameters=type_parameters,
                        module=module,
                        localns=localns,
                        recursion_guard=recursion_guard,
                    ),
                    AttributeAnnotation(
                        origin=NoneType,
                        arguments=[],
                    ),
                ],
            )

        case typing.Literal:
            return AttributeAnnotation(
                origin=Literal,
                arguments=list(get_args(annotation)),
            )

        case other:  # finally use whatever there was
            return AttributeAnnotation(
                origin=other,
                arguments=[
                    _resolve_attribute_annotation(
                        argument,
                        self_annotation=self_annotation,
                        type_parameters=type_parameters,
                        module=module,
                        localns=localns,
                        recursion_guard=recursion_guard,
                    )
                    for argument in get_args(other)
                ],
            )



from haiway.context import (
    Dependencies,
    Dependency,
    MissingContext,
    MissingDependency,
    MissingState,
    ScopeMetrics,
    ctx,
)
from haiway.helpers import (
    asynchronous,
    auto_retry,
    cached,
    throttle,
    with_timeout,
)
from haiway.state import Structure
from haiway.types import (
    MISSING,
    Missing,
    frozenlist,
    is_missing,
    not_missing,
    when_missing,
)
from haiway.utils import (
    AsyncQueue,
    always,
    async_always,
    async_noop,
    freeze,
    getenv_bool,
    getenv_float,
    getenv_int,
    getenv_str,
    load_env,
    mimic_function,
    noop,
    setup_logging,
)

__all__ = [
    "always",
    "async_always",
    "async_noop",
    "asynchronous",
    "AsyncQueue",
    "auto_retry",
    "cached",
    "ctx",
    "Dependencies",
    "Dependency",
    "freeze",
    "frozenlist",
    "getenv_bool",
    "getenv_float",
    "getenv_int",
    "getenv_str",
    "is_missing",
    "load_env",
    "mimic_function",
    "Missing",
    "MISSING",
    "MissingContext",
    "MissingDependency",
    "MissingState",
    "noop",
    "not_missing",
    "ScopeMetrics",
    "setup_logging",
    "Structure",
    "throttle",
    "when_missing",
    "with_timeout",
]



from asyncio import Future, gather, get_event_loop
from collections.abc import Callable
from contextvars import ContextVar, Token
from copy import copy
from itertools import chain
from logging import DEBUG, ERROR, INFO, WARNING, Logger, getLogger
from time import monotonic
from types import TracebackType
from typing import Any, Self, cast, final, overload
from uuid import uuid4

from haiway.state import Structure
from haiway.utils import freeze

__all__ = [
    "ScopeMetrics",
    "MetricsContext",
]


@final
class ScopeMetrics:
    def __init__(
        self,
        *,
        trace_id: str | None,
        scope: str,
        logger: Logger | None,
    ) -> None:
        self.trace_id: str = trace_id or uuid4().hex
        self._label: str = f"{self.trace_id}|{scope}" if scope else self.trace_id
        self._logger: Logger = logger or getLogger(name=scope)
        self._metrics: dict[type[Structure], Structure] = {}
        self._nested: list[ScopeMetrics] = []
        self._timestamp: float = monotonic()
        self._completed: Future[float] = get_event_loop().create_future()

        freeze(self)

    def __del__(self) -> None:
        self._complete()  # ensure completion on deinit

    def __str__(self) -> str:
        return self._label

    def metrics(
        self,
        *,
        merge: Callable[[Structure, Structure], Structure] = lambda lhs, rhs: lhs,
    ) -> list[Structure]:
        metrics: dict[type[Structure], Structure] = copy(self._metrics)
        for metric in chain.from_iterable(nested.metrics(merge=merge) for nested in self._nested):
            metric_type: type[Structure] = type(metric)
            if current := metrics.get(metric_type):
                metrics[metric_type] = merge(current, metric)

            else:
                metrics[metric_type] = metric

        return list(metrics.values())

    @overload
    def read[Metric: Structure](
        self,
        metric: type[Metric],
        /,
    ) -> Metric | None: ...

    @overload
    def read[Metric: Structure](
        self,
        metric: type[Metric],
        /,
        default: Metric,
    ) -> Metric: ...

    def read[Metric: Structure](
        self,
        metric: type[Metric],
        /,
        default: Metric | None = None,
    ) -> Metric | None:
        return cast(Metric | None, self._metrics.get(metric, default))

    def record[Metric: Structure](
        self,
        metric: Metric,
        /,
        *,
        merge: Callable[[Metric, Metric], Metric] = lambda lhs, rhs: rhs,
    ) -> None:
        assert not self._completed.done(), "Can't record using completed metrics scope"  # nosec: B101
        metric_type: type[Metric] = type(metric)
        if current := self._metrics.get(metric_type):
            self._metrics[metric_type] = merge(cast(Metric, current), metric)

        else:
            self._metrics[metric_type] = metric

    @property
    def completed(self) -> bool:
        return self._completed.done() and all(nested.completed for nested in self._nested)

    @property
    def time(self) -> float:
        if self._completed.done():
            return self._completed.result()

        else:
            return monotonic() - self._timestamp

    async def wait(self) -> None:
        await gather(
            self._completed,
            *[nested.wait() for nested in self._nested],
            return_exceptions=False,
        )

    def _complete(self) -> None:
        if self._completed.done():
            return  # already completed

        self._completed.set_result(monotonic() - self._timestamp)

    def scope(
        self,
        name: str,
        /,
    ) -> Self:
        nested: Self = self.__class__(
            scope=name,
            logger=self._logger,
            trace_id=self.trace_id,
        )
        self._nested.append(nested)
        return nested

    def log(
        self,
        level: int,
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        self._logger.log(
            level,
            f"[{self}] {message}",
            *args,
            exc_info=exception,
        )


@final
class MetricsContext:
    _context = ContextVar[ScopeMetrics]("MetricsContext")

    @classmethod
    def scope(
        cls,
        name: str,
        /,
        *,
        trace_id: str | None = None,
        logger: Logger | None = None,
    ) -> Self:
        try:
            context: ScopeMetrics = cls._context.get()
            if trace_id is None or context.trace_id == trace_id:
                return cls(context.scope(name))

            else:
                return cls(
                    ScopeMetrics(
                        trace_id=trace_id,
                        scope=name,
                        logger=logger or context._logger,  # pyright: ignore[reportPrivateUsage]
                    )
                )
        except LookupError:  # create metrics scope when missing yet
            return cls(
                ScopeMetrics(
                    trace_id=trace_id,
                    scope=name,
                    logger=logger,
                )
            )

    @classmethod
    def record[Metric: Structure](
        cls,
        metric: Metric,
        /,
        *,
        merge: Callable[[Metric, Metric], Metric] = lambda lhs, rhs: rhs,
    ) -> None:
        try:  # catch exceptions - we don't wan't to blow up on metrics
            cls._context.get().record(metric, merge=merge)

        except Exception as exc:
            cls.log_error(
                "Failed to record metric: %s",
                type(metric).__qualname__,
                exception=exc,
            )

    # - LOGS -

    @classmethod
    def log_error(
        cls,
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                ERROR,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                ERROR,
                message,
                *args,
                exc_info=exception,
            )

    @classmethod
    def log_warning(
        cls,
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                WARNING,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                WARNING,
                message,
                *args,
                exc_info=exception,
            )

    @classmethod
    def log_info(
        cls,
        message: str,
        /,
        *args: Any,
    ) -> None:
        try:
            cls._context.get().log(
                INFO,
                message,
                *args,
            )

        except LookupError:
            getLogger().log(
                INFO,
                message,
                *args,
            )

    @classmethod
    def log_debug(
        cls,
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        try:
            cls._context.get().log(
                DEBUG,
                message,
                *args,
                exception=exception,
            )

        except LookupError:
            getLogger().log(
                DEBUG,
                message,
                *args,
                exc_info=exception,
            )

    def __init__(
        self,
        metrics: ScopeMetrics,
    ) -> None:
        self._metrics: ScopeMetrics = metrics
        self._token: Token[ScopeMetrics] | None = None
        self._started: float | None = None
        self._finished: float | None = None

    def __enter__(self) -> None:
        assert (  # nosec: B101
            self._token is None and self._started is None
        ), "MetricsContext reentrance is not allowed"
        self._token = MetricsContext._context.set(self._metrics)
        self._started = monotonic()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        assert (  # nosec: B101
            self._token is not None and self._started is not None and self._finished is None
        ), "Unbalanced MetricsContext context enter/exit"
        self._finished = monotonic()
        MetricsContext._context.reset(self._token)
        self._token = None



from asyncio import (
    Task,
    current_task,
)
from collections.abc import (
    Callable,
    Coroutine,
)
from logging import Logger
from types import TracebackType
from typing import Any, final

from haiway.context.dependencies import Dependencies, Dependency
from haiway.context.metrics import MetricsContext, ScopeMetrics
from haiway.context.state import StateContext
from haiway.context.tasks import TaskGroupContext
from haiway.state import Structure
from haiway.utils import freeze

__all__ = [
    "ctx",
]


@final
class ScopeContext:
    def __init__(
        self,
        task_group: TaskGroupContext,
        state: StateContext,
        metrics: MetricsContext,
        completion: Callable[[ScopeMetrics], Coroutine[None, None, None]] | None,
    ) -> None:
        self._task_group: TaskGroupContext = task_group
        self._state: StateContext = state
        self._metrics: MetricsContext = metrics
        self._completion: Callable[[ScopeMetrics], Coroutine[None, None, None]] | None = completion

        freeze(self)

    def __enter__(self) -> None:
        assert self._completion is None, "Can't enter synchronous context with completion"  # nosec: B101

        self._state.__enter__()
        self._metrics.__enter__()

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._metrics.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._state.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

    async def __aenter__(self) -> None:
        self._state.__enter__()
        self._metrics.__enter__()
        await self._task_group.__aenter__()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._task_group.__aexit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._metrics.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        self._state.__exit__(
            exc_type=exc_type,
            exc_val=exc_val,
            exc_tb=exc_tb,
        )

        if completion := self._completion:
            await completion(self._metrics._metrics)  # pyright: ignore[reportPrivateUsage]


@final
class ctx:
    @staticmethod
    def scope(
        name: str,
        /,
        *state: Structure,
        logger: Logger | None = None,
        trace_id: str | None = None,
        completion: Callable[[ScopeMetrics], Coroutine[None, None, None]] | None = None,
    ) -> ScopeContext:
        """
        Access scope context with given parameters. When called within an existing context\
         it becomes nested with current context as its predecessor.

        Parameters
        ----------
        name: Value
            name of the scope context

        *state: Structure
            state propagated within the scope context, will be merged with current if any\
             by replacing current with provided on conflict

        logger: Logger | None
            logger used within the scope context, when not provided current logger will be used\
             if any, otherwise the logger with the scope name will be requested.

        trace_id: str | None = None
            tracing identifier included in logs produced within the scope context, when not\
             provided current identifier will be used if any, otherwise it random id will\
             be generated

        completion: Callable[[ScopeMetrics], Coroutine[None, None, None]] | None = None
            completion callback called on exit from the scope granting access to finished\
             scope metrics. Completion is called outside of the context when its metrics is\
             already finished. Make sure to avoid any long operations within the completion.

        Returns
        -------
        ScopeContext
            context object intended to enter context manager with it
        """

        return ScopeContext(
            task_group=TaskGroupContext(),
            metrics=MetricsContext.scope(
                name,
                logger=logger,
                trace_id=trace_id,
            ),
            state=StateContext.updated(state),
            completion=completion,
        )

    @staticmethod
    def updated(
        *state: Structure,
    ) -> StateContext:
        """
        Update scope context with given state. When called within an existing context\
         it becomes nested with current context as its predecessor.

        Parameters
        ----------
        *state: Structure
            state propagated within the updated scope context, will be merged with current if any\
             by replacing current with provided on conflict

        Returns
        -------
        StateContext
            state part of context object intended to enter context manager with it
        """

        return StateContext.updated(state)

    @staticmethod
    def spawn[Result, **Arguments](
        function: Callable[Arguments, Coroutine[None, None, Result]],
        /,
        *args: Arguments.args,
        **kwargs: Arguments.kwargs,
    ) -> Task[Result]:
        """
        Spawn an async task within current scope context task group. When called outside of context\
         it will spawn detached task instead.

        Parameters
        ----------
        function: Callable[Arguments, Coroutine[None, None, Result]]
            function to be called within the task group

        *args: Arguments.args
            positional arguments passed to function call

        **kwargs: Arguments.kwargs
            keyword arguments passed to function call

        Returns
        -------
        Task[Result]
            task for tracking function execution and result
        """

        return TaskGroupContext.run(function, *args, **kwargs)

    @staticmethod
    def cancel() -> None:
        """
        Cancel current asyncio task
        """

        if task := current_task():
            task.cancel()

        else:
            raise RuntimeError("Attempting to cancel context out of asyncio task")

    @staticmethod
    async def dependency[DependencyType: Dependency](
        dependency: type[DependencyType],
        /,
    ) -> DependencyType:
        """
        Access current dependency by its type.

        Parameters
        ----------
        dependency: type[DependencyType]
            type of requested dependency

        Returns
        -------
        DependencyType
            resolved dependency instance
        """

        return await Dependencies.dependency(dependency)

    @staticmethod
    def state[StateType: Structure](
        state: type[StateType],
        /,
        default: StateType | None = None,
    ) -> StateType:
        """
        Access current scope context state by its type. If there is no matching state defined\
         default value will be created if able, an exception will raise otherwise.

        Parameters
        ----------
        state: type[StateType]
            type of requested state

        Returns
        -------
        StateType
            resolved state instance
        """
        return StateContext.current(
            state,
            default=default,
        )

    @staticmethod
    def record[Metric: Structure](
        metric: Metric,
        /,
        merge: Callable[[Metric, Metric], Metric] = lambda lhs, rhs: rhs,
    ) -> None:
        """
        Record metric within current scope context.

        Parameters
        ----------
        metric: MetricType
            value of metric to be recorded

        merge: Callable[[MetricType, MetricType], MetricType] = lambda lhs, rhs: rhs
            merge method used on to resolve conflicts when a metric of the same type\
             was already recorded. When not provided value will be override current if any.

        Returns
        -------
        None
        """

        MetricsContext.record(
            metric,
            merge=merge,
        )

    @staticmethod
    def log_error(
        message: str,
        /,
        *args: Any,
        exception: BaseException | None = None,
    ) -> None:
        """
        Log using ERROR level within current scope context. When there is no current scope\
         root logger will be used without additional details.

        Parameters
        ----------
        message: str
            message to be written to log

        *args: Any
            message format arguments

        exception: BaseException | None = None
            exception associated with log, when provided full stack trace will be recorded

        Returns
        -------
        None
        """

        MetricsContext.log_error(
            message,
            *args,
            exception=exception,
        )

    @staticmethod
    def log_warning(
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        """
        Log using WARNING level within current scope context. When there is no current scope\
         root logger will be used without additional details.

        Parameters
        ----------
        message: str
            message to be written to log

        *args: Any
            message format arguments

        exception: BaseException | None = None
            exception associated with log, when provided full stack trace will be recorded

        Returns
        -------
        None
        """

        MetricsContext.log_warning(
            message,
            *args,
            exception=exception,
        )

    @staticmethod
    def log_info(
        message: str,
        /,
        *args: Any,
    ) -> None:
        """
        Log using INFO level within current scope context. When there is no current scope\
         root logger will be used without additional details.

        Parameters
        ----------
        message: str
            message to be written to log

        *args: Any
            message format arguments

        Returns
        -------
        None
        """

        MetricsContext.log_info(
            message,
            *args,
        )

    @staticmethod
    def log_debug(
        message: str,
        /,
        *args: Any,
        exception: Exception | None = None,
    ) -> None:
        """
        Log using DEBUG level within current scope context. When there is no current scope\
         root logger will be used without additional details.

        Parameters
        ----------
        message: str
            message to be written to log

        *args: Any
            message format arguments

        exception: BaseException | None = None
            exception associated with log, when provided full stack trace will be recorded

        Returns
        -------
        None
        """

        MetricsContext.log_debug(
            message,
            *args,
            exception=exception,
        )



from haiway.state.attributes import AttributeAnnotation, attribute_annotations
from haiway.state.structure import Structure

__all__ = [
    "attribute_annotations",
    "AttributeAnnotation",
    "Structure",
]



from collections.abc import Callable
from copy import deepcopy
from types import GenericAlias
from typing import (
    Any,
    ClassVar,
    Generic,
    Self,
    TypeVar,
    cast,
    dataclass_transform,
    final,
    get_origin,
)
from weakref import WeakValueDictionary

from haiway.state.attributes import AttributeAnnotation, attribute_annotations
from haiway.state.validation import attribute_type_validator
from haiway.types.missing import MISSING, Missing

__all__ = [
    "Structure",
]


@final
class StructureAttribute[Value]:
    def __init__(
        self,
        annotation: AttributeAnnotation,
        default: Value | Missing,
        validator: Callable[[Any], Value],
    ) -> None:
        self.annotation: AttributeAnnotation = annotation
        self.default: Value | Missing = default
        self.validator: Callable[[Any], Value] = validator

    def validated(
        self,
        value: Any | Missing,
        /,
    ) -> Value:
        return self.validator(self.default if value is MISSING else value)


@dataclass_transform(
    kw_only_default=True,
    frozen_default=True,
    field_specifiers=(),
)
class StructureMeta(type):
    def __new__(
        cls,
        /,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        type_parameters: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Any:
        structure_type = type.__new__(
            cls,
            name,
            bases,
            namespace,
            **kwargs,
        )

        attributes: dict[str, StructureAttribute[Any]] = {}

        if bases:  # handle base class
            for key, annotation in attribute_annotations(
                structure_type,
                type_parameters=type_parameters,
            ).items():
                # do not include ClassVars and dunder items
                if ((get_origin(annotation) or annotation) is ClassVar) or key.startswith("__"):
                    continue

                attributes[key] = StructureAttribute(
                    annotation=annotation,
                    default=getattr(structure_type, key, MISSING),
                    validator=attribute_type_validator(annotation),
                )

        structure_type.__ATTRIBUTES__ = attributes  # pyright: ignore[reportAttributeAccessIssue]
        structure_type.__slots__ = frozenset(attributes.keys())  # pyright: ignore[reportAttributeAccessIssue]
        structure_type.__match_args__ = structure_type.__slots__  # pyright: ignore[reportAttributeAccessIssue]

        return structure_type


_types_cache: WeakValueDictionary[
    tuple[
        Any,
        tuple[Any, ...],
    ],
    Any,
] = WeakValueDictionary()


class Structure(metaclass=StructureMeta):
    """
    Base class for immutable data structures.
    """

    __ATTRIBUTES__: ClassVar[dict[str, StructureAttribute[Any]]]

    def __class_getitem__(
        cls,
        type_argument: tuple[type[Any], ...] | type[Any],
    ) -> type[Self]:
        assert Generic in cls.__bases__, "Can't specialize non generic type!"  # nosec: B101

        type_arguments: tuple[type[Any], ...]
        match type_argument:
            case [*arguments]:
                type_arguments = tuple(arguments)

            case argument:
                type_arguments = (argument,)

        if any(isinstance(argument, TypeVar) for argument in type_arguments):  # pyright: ignore[reportUnnecessaryIsInstance]
            # if we got unfinished type treat it as an alias instead of resolving
            return cast(type[Self], GenericAlias(cls, type_arguments))

        assert len(type_arguments) == len(  # nosec: B101
            cls.__type_params__
        ), "Type arguments count has to match type parameters count"

        if cached := _types_cache.get((cls, type_arguments)):
            return cached

        type_parameters: dict[str, Any] = {
            parameter.__name__: argument
            for (parameter, argument) in zip(
                cls.__type_params__ or (),
                type_arguments or (),
                strict=False,
            )
        }

        parameter_names: str = ",".join(
            getattr(
                argument,
                "__name__",
                str(argument),
            )
            for argument in type_arguments
        )
        name: str = f"{cls.__name__}[{parameter_names}]"
        bases: tuple[type[Self]] = (cls,)

        parametrized_type: type[Self] = StructureMeta.__new__(
            cls.__class__,
            name=name,
            bases=bases,
            namespace={"__module__": cls.__module__},
            type_parameters=type_parameters,
        )
        _types_cache[(cls, type_arguments)] = parametrized_type
        return parametrized_type

    def __init__(
        self,
        **kwargs: Any,
    ) -> None:
        for name, attribute in self.__ATTRIBUTES__.items():
            object.__setattr__(
                self,  # pyright: ignore[reportUnknownArgumentType]
                name,
                attribute.validated(
                    kwargs.get(
                        name,
                        MISSING,
                    ),
                ),
            )

    def updated(
        self,
        **kwargs: Any,
    ) -> Self:
        return self.__replace__(**kwargs)

    def as_dict(self) -> dict[str, Any]:
        return vars(self)

    def __str__(self) -> str:
        attributes: str = ", ".join([f"{key}: {value}" for key, value in vars(self).items()])
        return f"{self.__class__.__name__}({attributes})"

    def __repr__(self) -> str:
        return str(self)

    def __eq__(
        self,
        other: Any,
    ) -> bool:
        if not issubclass(other.__class__, self.__class__):
            return False

        return all(
            getattr(self, key, MISSING) == getattr(other, key, MISSING)
            for key in self.__ATTRIBUTES__.keys()
        )

    def __setattr__(
        self,
        name: str,
        value: Any,
    ) -> Any:
        raise AttributeError(
            f"Can't modify immutable structure {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be modified"
        )

    def __delattr__(
        self,
        name: str,
    ) -> None:
        raise AttributeError(
            f"Can't modify immutable structure {self.__class__.__qualname__},"
            f" attribute - '{name}' cannot be deleted"
        )

    def __copy__(self) -> Self:
        return self.__class__(**vars(self))

    def __deepcopy__(
        self,
        memo: dict[int, Any] | None,
    ) -> Self:
        copy: Self = self.__class__(
            **{
                key: deepcopy(
                    value,
                    memo,
                )
                for key, value in vars(self).items()
            }
        )
        return copy

    def __replace__(
        self,
        **kwargs: Any,
    ) -> Self:
        return self.__class__(
            **{
                **vars(self),
                **kwargs,
            }
        )
